@@ -3,7 +3,7 @@
 /***********************************************************************
 *  This code is part of GLPK (GNU Linear Programming Kit).
 *
-*  Copyright (C) 2015 Andrew Makhorin, Department for Applied
+*  Copyright (C) 2015-2017 Andrew Makhorin, Department for Applied
 *  Informatics, Moscow Aviation Institute, Moscow, Russia. All rights
 *  reserved. E-mail: <mao@gnu.org>.
 *
@@ -385,4 +385,214 @@ int spx_chuzr_harris(SPXLP *lp, int phase, const double beta[/*1+m*/],
 done: return p;
 }
 
+#if 1 /* 22/VI-2017 */
+/***********************************************************************
+*  spx_ls_eval_bp - determine penalty function break points
+*
+*  This routine determines break points of the penalty function (which
+*  is the sum of primal infeasibilities).
+*
+*  The parameters lp, beta, q, dq, tcol, and tol_piv have the same
+*  meaning as for the routine spx_chuzr_std (see above).
+*
+*  The routine stores the break-points determined to the array elements
+*  bp[1], ..., bp[nbp] in *arbitrary* order, where 0 <= nbp <= 2*m+1 is
+*  the number of break-points returned by the routine on exit. */
+
+int spx_ls_eval_bp(SPXLP *lp, const double beta[/*1+m*/],
+      int q, double dq, const double tcol[/*1+m*/], double tol_piv,
+      SPXBP bp[/*1+2*m+1*/])
+{     int m = lp->m;
+      int n = lp->n;
+      double *c = lp->c;
+      double *l = lp->l;
+      double *u = lp->u;
+      int *head = lp->head;
+      int i, k, nbp;
+      double s, alfa;
+      xassert(1 <= q && q <= n-m);
+      xassert(dq != 0.0);
+      s = (dq < 0.0 ? +1.0 : -1.0);
+      nbp = 0;
+      /* if chosen non-basic variable xN[q] is double-bounded, include
+       * it in the list, because it can cross its opposite bound */
+      k = head[m+q]; /* x[k] = xN[q] */
+      if (l[k] != -DBL_MAX && u[k] != +DBL_MAX)
+      {  nbp++;
+         bp[nbp].i = 0;
+         xassert(l[k] < u[k]); /* xN[q] cannot be fixed */
+         bp[nbp].teta = u[k] - l[k];
+         bp[nbp].dc = s;
+      }
+      /* build the list of all basic variables xB[i] that can cross
+       * their bound(s) for the ray parameter 0 <= teta < teta_max */
+      for (i = 1; i <= m; i++)
+      {  k = head[i]; /* x[k] = xB[i] */
+         xassert(l[k] <= u[k]);
+         /* determine alfa such that (delta xB[i]) = alfa * teta */
+         alfa = s * tcol[i];
+         if (alfa >= +tol_piv)
+         {  /* xB[i] increases on increasing teta */
+            if (l[k] == u[k])
+            {  /* xB[i] is fixed at lB[i] = uB[i] */
+               if (c[k] <= 0.0)
+               {  /* increasing xB[i] can cross its fixed value lB[i],
+                   * because currently xB[i] <= lB[i] */
+                  nbp++;
+                  bp[nbp].i = +i;
+                  bp[nbp].teta = (l[k] - beta[i]) / alfa;
+                  /* if xB[i] > lB[i] then cB[i] = +1 */
+                  bp[nbp].dc = +1.0 - c[k];
+               }
+            }
+            else
+            {  if (l[k] != -DBL_MAX && c[k] < 0.0)
+               {  /* increasing xB[i] can cross its lower bound lB[i],
+                   * because currently xB[i] < lB[i] */
+                  nbp++;
+                  bp[nbp].i = +i;
+                  bp[nbp].teta = (l[k] - beta[i]) / alfa;
+                  bp[nbp].dc = +1.0;
+               }
+               if (u[k] != +DBL_MAX && c[k] <= 0.0)
+               {  /* increasing xB[i] can cross its upper bound uB[i],
+                   * because currently xB[i] does not violate it */
+                  nbp++;
+                  bp[nbp].i = -i;
+                  bp[nbp].teta = (u[k] - beta[i]) / alfa;
+                  bp[nbp].dc = +1.0;
+               }
+            }
+         }
+         else if (alfa <= -tol_piv)
+         {  /* xB[i] decreases on increasing teta */
+            if (l[k] == u[k])
+            {  /* xB[i] is fixed at lB[i] = uB[i] */
+               if (c[k] >= 0.0)
+               {  /* decreasing xB[i] can cross its fixed value lB[i],
+                   * because currently xB[i] >= lB[i] */
+                  nbp++;
+                  bp[nbp].i = +i;
+                  bp[nbp].teta = (l[k] - beta[i]) / alfa;
+                  /* if xB[i] < lB[i] then cB[i] = -1 */
+                  bp[nbp].dc = -1.0 - c[k];
+               }
+            }
+            else
+            {  if (l[k] != -DBL_MAX && c[k] >= 0.0)
+               {  /* decreasing xB[i] can cross its lower bound lB[i],
+                   * because currently xB[i] does not violate it */
+                  nbp++;
+                  bp[nbp].i = +i;
+                  bp[nbp].teta = (l[k] - beta[i]) / alfa;
+                  bp[nbp].dc = -1.0;
+               }
+               if (u[k] != +DBL_MAX && c[k] > 0.0)
+               {  /* decreasing xB[i] can cross its upper bound uB[i],
+                   * because currently xB[i] > uB[i] */
+                  nbp++;
+                  bp[nbp].i = -i;
+                  bp[nbp].teta = (u[k] - beta[i]) / alfa;
+                  bp[nbp].dc = -1.0;
+               }
+            }
+         }
+         else
+         {  /* xB[i] does not depend on teta within a tolerance */
+            continue;
+         }
+         /* teta < 0 may happen only due to round-off errors when the
+          * current value of xB[i] is *close* to its (lower or upper)
+          * bound; in this case we replace teta by exact zero */
+         if (bp[nbp].teta < 0.0)
+            bp[nbp].teta = 0.0;
+      }
+      xassert(nbp <= 2*m+1);
+      return nbp;
+}
+#endif
+
+#if 1 /* 22/VI-2017 */
+/***********************************************************************
+*  spx_ls_select_bp - select and process penalty function break points
+*
+*  This routine selects a next portion of the penalty function break
+*  points and processes them.
+*
+*  On entry to the routine it is assumed that break points bp[1], ...,
+*  bp[num] are already processed, and slope is the penalty function
+*  slope to the right of the last processed break point bp[num].
+*  (Initially, when num = 0, slope should be specified as -fabs(d[q]),
+*  where d[q] is the reduced cost of chosen non-basic variable xN[q].)
+*
+*  The routine selects break points among bp[num+1], ..., bp[nbp], for
+*  which teta <= teta_lim, and moves these break points to the array
+*  elements bp[num+1], ..., bp[num1], where num <= num1 <= 2*m+1 is the
+*  new number of processed break points returned by the routine on
+*  exit. Then the routine sorts the break points by ascending teta and
+*  computes the change of the penalty function relative to its value at
+*  teta = 0.
+*
+*  On exit the routine also replaces the parameter slope with a new
+*  value that corresponds to the new last break-point bp[num1]. */
+
+#ifndef __WOE__
+static int fcmp(const void *v1, const void *v2)
+#else
+static int __cdecl fcmp(const void *v1, const void *v2)
+#endif
+{     const SPXBP *p1 = v1, *p2 = v2;
+      if (p1->teta < p2->teta)
+         return -1;
+      else if (p1->teta > p2->teta)
+         return +1;
+      else
+         return 0;
+}
+
+int spx_ls_select_bp(SPXLP *lp, const double tcol[/*1+m*/],
+      int nbp, SPXBP bp[/*1+m+m+1*/], int num, double *slope, double
+      teta_lim)
+{     int m = lp->m;
+      int i, t, num1;
+      double teta, dz;
+      xassert(0 <= num && num <= nbp && nbp <= m+m+1);
+      /* select a new portion of break points */
+      num1 = num;
+      for (t = num+1; t <= nbp; t++)
+      {  if (bp[t].teta <= teta_lim)
+         {  /* move break point to the beginning of the new portion */
+            num1++;
+            i = bp[num1].i, teta = bp[num1].teta, dz = bp[num1].dc;
+            bp[num1].i = bp[t].i, bp[num1].teta = bp[t].teta,
+            bp[num1].dc = bp[t].dc;
+            bp[t].i = i, bp[t].teta = teta, bp[t].dc = dz;
+         }
+      }
+      /* sort new break points bp[num+1], ..., bp[num1] by ascending
+       * the ray parameter teta */
+      if (num1 - num > 1)
+         qsort(&bp[num+1], num1 - num, sizeof(SPXBP), fcmp);
+      /* calculate the penalty function change at the new break points
+       * selected */
+      for (t = num+1; t <= num1; t++)
+      {  /* calculate the penalty function change relative to its value
+          * at break point bp[t-1] */
+         dz = (*slope) * (bp[t].teta - (t == 1 ? 0.0 : bp[t-1].teta));
+         /* calculate the penalty function change relative to its value
+          * at teta = 0 */
+         bp[t].dz = (t == 1 ? 0.0 : bp[t-1].dz) + dz;
+         /* calculate a new slope of the penalty function to the right
+          * of the current break point bp[t] */
+         i = (bp[t].i >= 0 ? bp[t].i : -bp[t].i);
+         xassert(0 <= i && i <= m);
+         if (i == 0)
+            *slope += fabs(1.0 * bp[t].dc);
+         else
+            *slope += fabs(tcol[i] * bp[t].dc);
+      }
+      return num1;
+}
+#endif
+
 /* eof */
