@@ -21,8 +21,8 @@
 *  along with GLPK. If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/
 
-#if 1 /* 26/V-2017 */
-#define PERTURB 1
+#if 1 /* 18/VII-2017 */
+#define SCALE_Z 1
 #endif
 
 #include "env.h"
@@ -33,18 +33,6 @@
 #include "spxchuzr.h"
 #include "spxprob.h"
 
-#define USE_AT 0
-/* 1 - use A in row-wise format
- * 0 - use N in row-wise format */
-
-#define EXCL 1
-/* 1 - exclude fixed non-basic variables
- * 0 - don't exclude variables */
-
-#define SHIFT 1
-/* 1 - shift bounds of variables toward zero
- * 0 - don't shift bounds of variables */
-
 #define CHECK_ACCURACY 0
 /* (for debugging) */
 
@@ -57,14 +45,16 @@ struct csa
       /* original optimization direction:
        * +1 - minimization
        * -1 - maximization */
+#if SCALE_Z
+      double fz;
+      /* factor used to scale original objective */
+#endif
       double *orig_c; /* double orig_c[1+n]; */
       /* copy of original objective coefficients */
-#if PERTURB
       double *orig_l; /* double orig_l[1+n]; */
       /* copy of original lower bounds */
       double *orig_u; /* double orig_u[1+n]; */
       /* copy of original upper bounds */
-#endif
       SPXAT *at;
       /* mxn-matrix A of constraint coefficients, in sparse row-wise
        * format (NULL if not used) */
@@ -77,14 +67,14 @@ struct csa
        * 1 - searching for primal feasible solution
        * 2 - searching for optimal solution */
       double *beta; /* double beta[1+m]; */
-      /* beta[i] is primal value of basic variable xB[i] */
+      /* beta[i] is a primal value of basic variable xB[i] */
       int beta_st;
       /* status of the vector beta:
        * 0 - undefined
        * 1 - just computed
        * 2 - updated */
       double *d; /* double d[1+n-m]; */
-      /* d[j] is reduced cost of non-basic variable xN[j] */
+      /* d[j] is a reduced cost of non-basic variable xN[j] */
       int d_st;
       /* status of the vector d:
        * 0 - undefined
@@ -106,6 +96,10 @@ struct csa
       FVS tcol; /* FVS tcol[1:m]; */
 #endif
       /* q-th (pivot) column of the simplex table */
+#if 1 /* 23/VI-2017 */
+      SPXBP *bp; /* SPXBP bp[1+2*m+1]; */
+      /* penalty function break points */
+#endif
       int p;
       /* xB[p] is a basic variable chosen to leave the basis;
        * p = 0 means that no basic variable reaches its bound;
@@ -120,18 +114,31 @@ struct csa
       FVS trow; /* FVS trow[1:n-m]; */
 #endif
       /* p-th (pivot) row of the simplex table */
+#if 0 /* 09/VII-2017 */
       double *work; /* double work[1+m]; */
       /* working array */
+#else
+      FVS work; /* FVS work[1:m]; */
+      /* working vector */
+#endif
       int p_stat, d_stat;
       /* primal and dual solution statuses */
       /*--------------------------------------------------------------*/
       /* control parameters (see struct glp_smcp) */
       int msg_lev;
       /* message level */
+#if 0 /* 23/VI-2017 */
       int harris;
       /* ratio test technique:
        * 0 - textbook ratio test
        * 1 - Harris' two pass ratio test */
+#else
+      int r_test;
+      /* ratio test technique:
+       * GLP_RT_STD  - textbook ratio test
+       * GLP_RT_HAR  - Harris' two pass ratio test
+       * GLP_RT_FLIP - long-step ratio test (only for phase I) */
+#endif
       double tol_bnd, tol_bnd1;
       /* primal feasibility tolerances */
       double tol_dj, tol_dj1;
@@ -143,7 +150,11 @@ struct csa
       int tm_lim;
       /* time limit, milliseconds */
       int out_frq;
+#if 0 /* 15/VII-2017 */
       /* display output frequency, iterations */
+#else
+      /* display output frequency, milliseconds */
+#endif
       int out_dly;
       /* display output delay, milliseconds */
       /*--------------------------------------------------------------*/
@@ -158,8 +169,21 @@ struct csa
        * jumps to its opposite bound) */
       int it_dpy;
       /* simplex iteration count at most recent display output */
+#if 1 /* 15/VII-2017 */
+      double tm_dpy;
+      /* time value at most recent display output */
+#endif
       int inv_cnt;
       /* basis factorization count since most recent display output */
+#if 1 /* 01/VII-2017 */
+      int degen;
+      /* count of successive degenerate iterations; this count is used
+       * to detect stalling */
+#endif
+#if 1 /* 23/VI-2017 */
+      int ns_cnt, ls_cnt;
+      /* normal and long-step iteration counts */
+#endif
 };
 
 /***********************************************************************
@@ -343,6 +367,7 @@ static int check_feas(struct csa *csa, int phase, double tol, double
 *  The routine returns the number of objective coefficients which were
 *  set to zero. */
 
+#if 0
 static int adjust_penalty(struct csa *csa, double tol, double tol1)
 {     SPXLP *lp = csa->lp;
       int m = lp->m;
@@ -378,6 +403,48 @@ static int adjust_penalty(struct csa *csa, double tol, double tol1)
       }
       return count;
 }
+#else
+static int adjust_penalty(struct csa *csa, int num, const int
+      ind[/*1+num*/], double tol, double tol1)
+{     SPXLP *lp = csa->lp;
+      int m = lp->m;
+      double *c = lp->c;
+      double *l = lp->l;
+      double *u = lp->u;
+      int *head = lp->head;
+      double *beta = csa->beta;
+      int i, k, t, cnt = 0;
+      double lk, uk, eps;
+      xassert(csa->phase == 1);
+      /* walk thru the specified list of basic variables */
+      for (t = 1; t <= num; t++)
+      {  i = ind[t];
+         xassert(1 <= i && i <= m);
+         k = head[i]; /* x[k] = xB[i] */
+         if (c[k] < 0.0)
+         {  /* x[k] violates its original lower bound */
+            lk = l[k];
+            xassert(lk != -DBL_MAX);
+            eps = tol + tol1 * (lk >= 0.0 ? +lk : -lk);
+            if (beta[i] >= lk - eps)
+            {  /* however, violation is close to zero */
+               c[k] = 0.0, cnt++;
+            }
+         }
+         else if (c[k] > 0.0)
+         {  /* x[k] violates its original upper bound */
+            uk = u[k];
+            xassert(uk != +DBL_MAX);
+            eps = tol + tol1 * (uk >= 0.0 ? +uk : -uk);
+            if (beta[i] <= uk + eps)
+            {  /* however, violation is close to zero */
+               c[k] = 0.0, cnt++;
+            }
+         }
+      }
+      return cnt;
+}
+#endif
 
 #if CHECK_ACCURACY
 /***********************************************************************
@@ -522,10 +589,21 @@ static int choose_pivot(struct csa *csa)
       double *d = csa->d;
       SPXSE *se = csa->se;
       int *list = csa->list;
+#if 0 /* 09/VII-2017 */
       double *tcol = csa->work;
+#else
+      double *tcol = csa->work.vec;
+#endif
       double tol_piv = csa->tol_piv;
       int try, nnn, /*i,*/ p, p_flag, q, t;
       double big, /*temp,*/ best_ratio;
+#if 1 /* 23/VI-2017 */
+      double *c = lp->c;
+      int *head = lp->head;
+      SPXBP *bp = csa->bp;
+      int nbp, t_best, ret, k;
+      double dz_best;
+#endif
       xassert(csa->beta_st);
       xassert(csa->d_st);
 more: /* initial number of eligible non-basic variables */
@@ -533,7 +611,11 @@ more: /* initial number of eligible non-basic variables */
       /* nothing has been chosen so far */
       csa->q = 0;
       best_ratio = 0.0;
+#if 0 /* 23/VI-2017 */
       try = 0;
+#else
+      try = ret = 0;
+#endif
 try:  /* choose non-basic variable xN[q] */
       xassert(nnn > 0);
       try++;
@@ -563,7 +645,106 @@ try:  /* choose non-basic variable xN[q] */
       big = 1.0;
 #endif
       /* choose basic variable xB[p] */
+#if 1 /* 23/VI-2017 */
+      if (csa->phase == 1 && csa->r_test == GLP_RT_FLIP && try <= 2)
+      {  /* long-step ratio test */
+         int t, num, num1;
+         double slope, teta_lim;
+         /* determine penalty function break points */
+         nbp = spx_ls_eval_bp(lp, beta, q, d[q], tcol, tol_piv, bp);
+         if (nbp < 2)
+            goto skip;
+         /* set initial slope */
+         slope = - fabs(d[q]);
+         /* estimate initial teta_lim */
+         teta_lim = DBL_MAX;
+         for (t = 1; t <= nbp; t++)
+         {  if (teta_lim > bp[t].teta)
+               teta_lim = bp[t].teta;
+         }
+         xassert(teta_lim >= 0.0);
+         if (teta_lim < 1e-3)
+            teta_lim = 1e-3;
+         /* nothing has been chosen so far */
+         t_best = 0, dz_best = 0.0, num = 0;
+         /* choose appropriate break point */
+         while (num < nbp)
+         {  /* select and process a new portion of break points */
+            num1 = spx_ls_select_bp(lp, tcol, nbp, bp, num, &slope,
+               teta_lim);
+            for (t = num+1; t <= num1; t++)
+            {  int i = (bp[t].i >= 0 ? bp[t].i : -bp[t].i);
+               xassert(0 <= i && i <= m);
+               if (i == 0 || fabs(tcol[i]) / big >= MIN_RATIO)
+               {  if (dz_best > bp[t].dz)
+                     t_best = t, dz_best = bp[t].dz;
+               }
+#if 0
+               if (i == 0)
+               {  /* do not consider further break points beyond this
+                   * point, where xN[q] reaches its opposite bound;
+                   * in principle (see spx_ls_eval_bp), this break
+                   * point should be the last one, however, due to
+                   * round-off errors there may be other break points
+                   * with the same teta beyond this one */
+                  slope = +1.0;
+               }
+#endif
+            }
+            if (slope > 0.0)
+            {  /* penalty function starts increasing */
+               break;
+            }
+            /* penalty function continues decreasing */
+            num = num1;
+            teta_lim += teta_lim;
+         }
+         if (dz_best == 0.0)
+            goto skip;
+         /* the choice has been made */
+         xassert(1 <= t_best && t_best <= num1);
+         if (t_best == 1)
+         {  /* the very first break point was chosen; it is reasonable
+             * to use the short-step ratio test */
+            goto skip;
+         }
+         csa->q = q;
+         memcpy(&csa->tcol.vec[1], &tcol[1], m * sizeof(double));
+         fvs_gather_vec(&csa->tcol, DBL_EPSILON);
+         if (bp[t_best].i == 0)
+         {  /* xN[q] goes to its opposite bound */
+            csa->p = -1;
+            csa->p_flag = 0;
+            best_ratio = 1.0;
+         }
+         else if (bp[t_best].i > 0)
+         {  /* xB[p] leaves the basis and goes to its lower bound */
+            csa->p = + bp[t_best].i;
+            xassert(1 <= csa->p && csa->p <= m);
+            csa->p_flag = 0;
+            best_ratio = fabs(tcol[csa->p]) / big;
+         }
+         else
+         {  /* xB[p] leaves the basis and goes to its upper bound */
+            csa->p = - bp[t_best].i;
+            xassert(1 <= csa->p && csa->p <= m);
+            csa->p_flag = 1;
+            best_ratio = fabs(tcol[csa->p]) / big;
+         }
+#if 0
+         xprintf("num1 = %d; t_best = %d; dz = %g\n", num1, t_best,
+            bp[t_best].dz);
+#endif
+         ret = 1;
+         goto done;
+skip:    ;
+      }
+#endif
+#if 0 /* 23/VI-2017 */
       if (!csa->harris)
+#else
+      if (csa->r_test == GLP_RT_STD)
+#endif
       {  /* textbook ratio test */
          p = spx_chuzr_std(lp, csa->phase, beta, q,
             d[q] < 0.0 ? +1. : -1., tcol, &p_flag, tol_piv,
@@ -630,11 +811,35 @@ done: /* the choice has been made */
          }
       }
 #endif
+#if 0 /* 23/VI-2017 */
       return 0;
+#else /* FIXME */
+      if (ret)
+      {  /* invalidate dual basic solution components */
+         csa->d_st = 0;
+         /* change penalty function coefficients at basic variables for
+          * all break points preceding the chosen one */
+         for (t = 1; t < t_best; t++)
+         {  int i = (bp[t].i >= 0 ? bp[t].i : -bp[t].i);
+            xassert(0 <= i && i <= m);
+            if (i == 0)
+            {  /* xN[q] crosses its opposite bound */
+               xassert(1 <= csa->q && csa->q <= n-m);
+               k = head[m+csa->q];
+            }
+            else
+            {  /* xB[i] crosses its (lower or upper) bound */
+               k = head[i]; /* x[k] = xB[i] */
+            }
+            c[k] += bp[t].dc;
+            xassert(c[k] == 0.0 || c[k] == +1.0 || c[k] == -1.0);
+         }
+      }
+      return ret;
+#endif
 }
 #endif
 
-#if PERTURB
 /***********************************************************************
 *  play_bounds - play bounds of primal variables
 *
@@ -739,7 +944,39 @@ static void play_bounds(struct csa *csa, int all)
       }
       return;
 }
+
+static void remove_perturb(struct csa *csa)
+{     /* remove perturbation */
+      SPXLP *lp = csa->lp;
+      int m = lp->m;
+      int n = lp->n;
+      double *l = lp->l;
+      double *u = lp->u;
+      int *head = lp->head;
+      char *flag = lp->flag;
+      double *orig_l = csa->orig_l;
+      double *orig_u = csa->orig_u;
+      int j, k;
+      /* restore original bounds of variables */
+      memcpy(l, orig_l, (1+n) * sizeof(double));
+      memcpy(u, orig_u, (1+n) * sizeof(double));
+      /* adjust flags of fixed non-basic variables, because in the
+       * perturbed problem such variables might be changed to double-
+       * bounded type */
+      for (j = 1; j <= n-m; j++)
+      {  k = head[m+j]; /* x[k] = xN[j] */
+         if (l[k] == u[k])
+            flag[j] = 0;
+      }
+      /* removing perturbation changes primal solution components */
+      csa->phase = csa->beta_st = 0;
+#if 1
+      if (csa->msg_lev >= GLP_MSG_ALL)
+         xprintf("Removing LP perturbation [%d]...\n",
+            csa->it_cnt);
 #endif
+      return;
+}
 
 /***********************************************************************
 *  sum_infeas - compute sum of primal infeasibilities
@@ -784,28 +1021,47 @@ static double sum_infeas(SPXLP *lp, const double beta[/*1+m*/])
 
 static void display(struct csa *csa, int spec)
 {     int nnn, k;
-      double obj, sum, *save;
+      double obj, sum, *save, *save1;
+#if 1 /* 15/VII-2017 */
+      double tm_cur;
+#endif
       /* check if the display output should be skipped */
       if (csa->msg_lev < GLP_MSG_ON) goto skip;
+#if 1 /* 15/VII-2017 */
+      tm_cur = xtime();
+#endif
       if (csa->out_dly > 0 &&
+#if 0 /* 15/VII-2017 */
          1000.0 * xdifftime(xtime(), csa->tm_beg) < csa->out_dly)
+#else
+         1000.0 * xdifftime(tm_cur, csa->tm_beg) < csa->out_dly)
+#endif
          goto skip;
       if (csa->it_cnt == csa->it_dpy) goto skip;
+#if 0 /* 15/VII-2017 */
       if (!spec && csa->it_cnt % csa->out_frq != 0) goto skip;
+#else
+      if (!spec &&
+         1000.0 * xdifftime(tm_cur, csa->tm_dpy) < csa->out_frq)
+         goto skip;
+#endif
       /* compute original objective value */
       save = csa->lp->c;
       csa->lp->c = csa->orig_c;
       obj = csa->dir * spx_eval_obj(csa->lp, csa->beta);
       csa->lp->c = save;
+#if SCALE_Z
+      obj *= csa->fz;
+#endif
       /* compute sum of (scaled) primal infeasibilities */
-#if 0 /* 27/V-2017 */
+#if 1 /* 01/VII-2017 */
       save = csa->lp->l;
       save1 = csa->lp->u;
       csa->lp->l = csa->orig_l;
       csa->lp->u = csa->orig_u;
 #endif
       sum = sum_infeas(csa->lp, csa->beta);
-#if 0 /* 27/V-2017 */
+#if 1 /* 01/VII-2017 */
       csa->lp->l = save;
       csa->lp->u = save1;
 #endif
@@ -832,8 +1088,20 @@ static void display(struct csa *csa, int spec)
          xprintf(" %d", csa->inv_cnt);
          csa->inv_cnt = 0;
       }
+#if 1 /* 23/VI-2017 */
+      if (csa->phase == 1 && csa->r_test == GLP_RT_FLIP)
+      {  /*xprintf("   %d,%d", csa->ns_cnt, csa->ls_cnt);*/
+         if (csa->ns_cnt + csa->ls_cnt)
+            xprintf(" %d%%",
+               (100 * csa->ls_cnt) / (csa->ns_cnt + csa->ls_cnt));
+         csa->ns_cnt = csa->ls_cnt = 0;
+      }
+#endif
       xprintf("\n");
       csa->it_dpy = csa->it_cnt;
+#if 1 /* 15/VII-2017 */
+      csa->tm_dpy = tm_cur;
+#endif
 skip: return;
 }
 
@@ -872,17 +1140,23 @@ static int primal_simplex(struct csa *csa)
       double *tcol = csa->tcol;
       double *trow = csa->trow;
 #endif
+#if 0 /* 09/VII-2017 */
       double *pi = csa->work;
       double *rho = csa->work;
+#else
+      double *pi = csa->work.vec;
+      double *rho = csa->work.vec;
+#endif
       int msg_lev = csa->msg_lev;
       double tol_bnd = csa->tol_bnd;
       double tol_bnd1 = csa->tol_bnd1;
       double tol_dj = csa->tol_dj;
       double tol_dj1 = csa->tol_dj1;
+      int perturb = -1;
+      /* -1 = perturbation is not used, but enabled
+       *  0 = perturbation is not used and disabled
+       * +1 = perturbation is being used */
       int j, refct, ret;
-#if PERTURB
-      int perturb = 1;
-#endif
 loop: /* main loop starts here */
       /* compute factorization of the basis matrix */
       if (!lp->valid)
@@ -938,42 +1212,45 @@ loop: /* main loop starts here */
          }
          /* make sure that the current basic solution remains primal
           * feasible (or pseudo-feasible on phase I) */
-#if PERTURB
-         if (perturb)
-            j = check_feas(csa, csa->phase, 3.14 * tol_bnd, 3.14 *
-               tol_bnd1);
-         else
-            j = check_feas(csa, csa->phase, tol_bnd, tol_bnd1);
-         if (j)
-#else
-         if (check_feas(csa, csa->phase, tol_bnd, tol_bnd1))
+         if (perturb <= 0)
+         {  if (check_feas(csa, csa->phase, tol_bnd, tol_bnd1))
+            {  /* excessive bound violations due to round-off errors */
+#if 1 /* 01/VII-2017 */
+               if (perturb < 0)
+               {  if (msg_lev >= GLP_MSG_ALL)
+                     xprintf("Perturbing LP to avoid instability [%d].."
+                        ".\n", csa->it_cnt);
+                  perturb = 1;
+                  goto loop;
+               }
 #endif
-         {  /* excessive bound violations due to round-off errors */
-            if (msg_lev >= GLP_MSG_ERR)
-               xprintf("Warning: numerical instability (primal simplex,"
-                  " phase %s)\n", csa->phase == 1 ? "I" : "II");
-            /* restart the search */
-            lp->valid = 0;
-            csa->phase = 0;
-            goto loop;
+               if (msg_lev >= GLP_MSG_ERR)
+                  xprintf("Warning: numerical instability (primal simpl"
+                     "ex, phase %s)\n", csa->phase == 1 ? "I" : "II");
+               /* restart the search */
+               lp->valid = 0;
+               csa->phase = 0;
+               goto loop;
+            }
+            if (csa->phase == 1)
+            {  int i, cnt;
+               for (i = 1; i <= m; i++)
+                  csa->tcol.ind[i] = i;
+               cnt = adjust_penalty(csa, m, csa->tcol.ind,
+                  0.99 * tol_bnd, 0.99 * tol_bnd1);
+               if (cnt)
+               {  /*xprintf("*** cnt = %d\n", cnt);*/
+                  csa->d_st = 0;
+               }
+            }
          }
-#if PERTURB
-         /* check_feas guarantees that all beta[i] are feasible within
-          * a tolerance */
-         if (perturb)
+         else
+         {  /* FIXME */
             play_bounds(csa, 1);
-#endif
+         }
       }
       /* at this point the search phase is determined */
       xassert(csa->phase == 1 || csa->phase == 2);
-      if (csa->phase == 1)
-      {  /* adjust penalty function coefficients */
-         if (adjust_penalty(csa, tol_bnd, tol_bnd1))
-         {  /* some coefficients were changed, so invalidate reduced
-             * costs of non-basic variables */
-            csa->d_st = 0;
-         }
-      }
       /* compute reduced costs of non-basic variables d = (d[j]) */
       if (!csa->d_st)
       {  spx_eval_pi(lp, pi);
@@ -994,16 +1271,11 @@ loop: /* main loop starts here */
 #endif
       /* check if the iteration limit has been exhausted */
       if (csa->it_cnt - csa->it_beg >= csa->it_lim)
-      {
-#if PERTURB
-         if (perturb)
+      {  if (perturb > 0)
          {  /* remove perturbation */
+            remove_perturb(csa);
             perturb = 0;
-            memcpy(csa->lp->l, csa->orig_l, (1+n) * sizeof(double));
-            memcpy(csa->lp->u, csa->orig_u, (1+n) * sizeof(double));
-            csa->phase = csa->beta_st = 0;
          }
-#endif
          if (csa->beta_st != 1)
             csa->beta_st = 0;
          if (csa->d_st != 1)
@@ -1020,16 +1292,11 @@ loop: /* main loop starts here */
       }
       /* check if the time limit has been exhausted */
       if (1000.0 * xdifftime(xtime(), csa->tm_beg) >= csa->tm_lim)
-      {
-#if PERTURB
-         if (perturb)
+      {  if (perturb > 0)
          {  /* remove perturbation */
+            remove_perturb(csa);
             perturb = 0;
-            memcpy(csa->lp->l, csa->orig_l, (1+n) * sizeof(double));
-            memcpy(csa->lp->u, csa->orig_u, (1+n) * sizeof(double));
-            csa->phase = csa->beta_st = 0;
          }
-#endif
          if (csa->beta_st != 1)
             csa->beta_st = 0;
          if (csa->d_st != 1)
@@ -1059,21 +1326,11 @@ loop: /* main loop starts here */
       }
       /* check for optimality */
       if (csa->num == 0)
-      {
-#if PERTURB
-         if (perturb && csa->phase == 2)
+      {  if (perturb > 0 && csa->phase == 2)
          {  /* remove perturbation */
+            remove_perturb(csa);
             perturb = 0;
-            memcpy(csa->lp->l, csa->orig_l, (1+n) * sizeof(double));
-            memcpy(csa->lp->u, csa->orig_u, (1+n) * sizeof(double));
-            csa->phase = csa->beta_st = 0;
-#if 0
-            if (msg_lev >= GLP_MSG_ALL)
-               xprintf("Perturbation removed at iteration %d\n",
-                  csa->it_cnt);
-#endif
          }
-#endif
          if (csa->beta_st != 1)
             csa->beta_st = 0;
          if (csa->d_st != 1)
@@ -1093,6 +1350,9 @@ loop: /* main loop starts here */
                   goto loop;
                }
                /* no feasible solution exists */
+#if 1 /* 09/VII-2017 */
+               /* FIXME: remove perturbation */
+#endif
                if (msg_lev >= GLP_MSG_ALL)
                   xprintf("LP HAS NO PRIMAL FEASIBLE SOLUTION\n");
                csa->p_stat = GLP_NOFEAS;
@@ -1111,6 +1371,7 @@ loop: /* main loop starts here */
          }
       }
       /* choose xN[q] and xB[p] */
+#if 0 /* 23/VI-2017 */
 #if 0 /* 17/III-2016 */
       choose_pivot(csa);
 #else
@@ -1118,19 +1379,25 @@ loop: /* main loop starts here */
       {  lp->valid = 0;
          goto loop;
       }
+#endif
+#else
+      ret = choose_pivot(csa);
+      if (ret < 0)
+      {  lp->valid = 0;
+         goto loop;
+      }
+      if (ret == 0)
+         csa->ns_cnt++;
+      else
+         csa->ls_cnt++;
 #endif
       /* check for unboundedness */
       if (csa->p == 0)
-      {
-#if PERTURB
-         if (perturb)
+      {  if (perturb > 0)
          {  /* remove perturbation */
+            remove_perturb(csa);
             perturb = 0;
-            memcpy(csa->lp->l, csa->orig_l, (1+n) * sizeof(double));
-            memcpy(csa->lp->u, csa->orig_u, (1+n) * sizeof(double));
-            csa->phase = csa->beta_st = 0;
          }
-#endif
          if (csa->beta_st != 1)
             csa->beta_st = 0;
          if (csa->d_st != 1)
@@ -1158,6 +1425,45 @@ loop: /* main loop starts here */
                xassert(csa != csa);
          }
       }
+#if 1 /* 01/VII-2017 */
+      /* check for stalling */
+      if (csa->p > 0)
+      {  int k;
+         xassert(1 <= csa->p && csa->p <= m);
+         k = head[csa->p]; /* x[k] = xB[p] */
+         if (lp->l[k] != lp->u[k])
+         {  if (csa->p_flag)
+            {  /* xB[p] goes to its upper bound */
+               xassert(lp->u[k] != +DBL_MAX);
+               if (fabs(beta[csa->p] - lp->u[k]) >= 1e-6)
+               {  csa->degen = 0;
+                  goto skip1;
+               }
+            }
+            else if (lp->l[k] == -DBL_MAX)
+            {  /* unusual case */
+               goto skip1;
+            }
+            else
+            {  /* xB[p] goes to its lower bound */
+               xassert(lp->l[k] != -DBL_MAX);
+               if (fabs(beta[csa->p] - lp->l[k]) >= 1e-6)
+               {  csa->degen = 0;
+                  goto skip1;
+               }
+            }
+            /* degenerate iteration has been detected */
+            csa->degen++;
+            if (perturb < 0 && csa->degen >= 200)
+            {  if (msg_lev >= GLP_MSG_ALL)
+                  xprintf("Perturbing LP to avoid stalling [%d]...\n",
+                     csa->it_cnt);
+               perturb = 1;
+            }
+skip1:      ;
+         }
+      }
+#endif
       /* update values of basic variables for adjacent basis */
 #if 0 /* 11/VI-2017 */
       spx_update_beta(lp, beta, csa->p, csa->p_flag, csa->q, tcol);
@@ -1199,6 +1505,10 @@ loop: /* main loop starts here */
 #endif
       /* update reduced costs of non-basic variables for adjacent
        * basis */
+#if 1 /* 23/VI-2017 */
+      /* dual solution may be invalidated due to long step */
+      if (csa->d_st)
+#endif
 #if 0 /* 11/VI-2017 */
       if (spx_update_d(lp, d, csa->p, csa->q, trow, tcol) <= 1e-9)
 #else
@@ -1253,9 +1563,25 @@ skip: /* change current basis header to adjacent one */
       /* and update factorization of the basis matrix */
       if (csa->p > 0)
          spx_update_invb(lp, csa->p, head[csa->p]);
-#if PERTURB
-      if (perturb && csa->beta_st)
+#if 1
+      if (perturb <= 0)
+      {  if (csa->phase == 1)
+         {  int cnt;
+            /* adjust penalty function coefficients */
+            cnt = adjust_penalty(csa, csa->tcol.nnz, csa->tcol.ind,
+               0.99 * tol_bnd, 0.99 * tol_bnd1);
+            if (cnt)
+            {  /* some coefficients were changed, so invalidate reduced
+                * costs of non-basic variables */
+               /*xprintf("... cnt = %d\n", cnt);*/
+               csa->d_st = 0;
+            }
+         }
+      }
+      else
+      {  /* FIXME */
          play_bounds(csa, 0);
+      }
 #endif
       /* simplex iteration complete */
       csa->it_cnt++;
@@ -1279,20 +1605,20 @@ int spx_primal(glp_prob *P, const glp_smcp *parm)
 {     /* driver to the primal simplex method */
       struct csa csa_, *csa = &csa_;
       SPXLP lp;
-#if USE_AT
       SPXAT at;
-#else
       SPXNT nt;
-#endif
       SPXSE se;
       int ret, *map, *daeh;
+#if SCALE_Z
+      int i, j, k;
+#endif
       /* build working LP and its initial basis */
       memset(csa, 0, sizeof(struct csa));
       csa->lp = &lp;
-      spx_init_lp(csa->lp, P, EXCL);
+      spx_init_lp(csa->lp, P, parm->excl);
       spx_alloc_lp(csa->lp);
       map = talloc(1+P->m+P->n, int);
-      spx_build_lp(csa->lp, P, EXCL, SHIFT, map);
+      spx_build_lp(csa->lp, P, parm->excl, parm->shift, map);
       spx_build_basis(csa->lp, P, map);
       switch (P->dir)
       {  case GLP_MIN:
@@ -1304,28 +1630,50 @@ int spx_primal(glp_prob *P, const glp_smcp *parm)
          default:
             xassert(P != P);
       }
+#if SCALE_Z
+      csa->fz = 0.0;
+      for (k = 1; k <= csa->lp->n; k++)
+      {  double t = fabs(csa->lp->c[k]);
+         if (csa->fz < t)
+            csa->fz = t;
+      }
+      if (csa->fz <= 1000.0)
+         csa->fz = 1.0;
+      else
+         csa->fz /= 1000.0;
+      /*xprintf("csa->fz = %g\n", csa->fz);*/
+      for (k = 0; k <= csa->lp->n; k++)
+         csa->lp->c[k] /= csa->fz;
+#endif
       csa->orig_c = talloc(1+csa->lp->n, double);
       memcpy(csa->orig_c, csa->lp->c, (1+csa->lp->n) * sizeof(double));
-#if PERTURB
+#if 1 /*PERTURB*/
       csa->orig_l = talloc(1+csa->lp->n, double);
       memcpy(csa->orig_l, csa->lp->l, (1+csa->lp->n) * sizeof(double));
       csa->orig_u = talloc(1+csa->lp->n, double);
       memcpy(csa->orig_u, csa->lp->u, (1+csa->lp->n) * sizeof(double));
-#endif
-#if USE_AT
-      /* build matrix A in row-wise format */
-      csa->at = &at;
-      csa->nt = NULL;
-      spx_alloc_at(csa->lp, csa->at);
-      spx_build_at(csa->lp, csa->at);
 #else
-      /* build matrix N in row-wise format for initial basis */
-      csa->at = NULL;
-      csa->nt = &nt;
-      spx_alloc_nt(csa->lp, csa->nt);
-      spx_init_nt(csa->lp, csa->nt);
-      spx_build_nt(csa->lp, csa->nt);
+      csa->orig_l = csa->orig_u = NULL;
 #endif
+      switch (parm->aorn)
+      {  case GLP_USE_AT:
+            /* build matrix A in row-wise format */
+            csa->at = &at;
+            csa->nt = NULL;
+            spx_alloc_at(csa->lp, csa->at);
+            spx_build_at(csa->lp, csa->at);
+            break;
+         case GLP_USE_NT:
+            /* build matrix N in row-wise format for initial basis */
+            csa->at = NULL;
+            csa->nt = &nt;
+            spx_alloc_nt(csa->lp, csa->nt);
+            spx_init_nt(csa->lp, csa->nt);
+            spx_build_nt(csa->lp, csa->nt);
+            break;
+         default:
+            xassert(parm != parm);
+      }
       /* allocate and initialize working components */
       csa->phase = 0;
       csa->beta = talloc(1+csa->lp->m, double);
@@ -1351,9 +1699,17 @@ int spx_primal(glp_prob *P, const glp_smcp *parm)
       fvs_alloc_vec(&csa->tcol, csa->lp->m);
       fvs_alloc_vec(&csa->trow, csa->lp->n-csa->lp->m);
 #endif
+#if 1 /* 23/VI-2017 */
+      csa->bp = NULL;
+#endif
+#if 0 /* 09/VII-2017 */
       csa->work = talloc(1+csa->lp->m, double);
+#else
+      fvs_alloc_vec(&csa->work, csa->lp->m);
+#endif
       /* initialize control parameters */
       csa->msg_lev = parm->msg_lev;
+#if 0 /* 23/VI-2017 */
       switch (parm->r_test)
       {  case GLP_RT_STD:
             csa->harris = 0;
@@ -1370,6 +1726,19 @@ int spx_primal(glp_prob *P, const glp_smcp *parm)
          default:
             xassert(parm != parm);
       }
+#else
+      switch (parm->r_test)
+      {  case GLP_RT_STD:
+         case GLP_RT_HAR:
+            break;
+         case GLP_RT_FLIP:
+            csa->bp = talloc(1+2*csa->lp->m+1, SPXBP);
+            break;
+         default:
+            xassert(parm != parm);
+      }
+      csa->r_test = parm->r_test;
+#endif
       csa->tol_bnd = parm->tol_bnd;
       csa->tol_bnd1 = .001 * parm->tol_bnd;
       csa->tol_dj = parm->tol_dj;
@@ -1383,7 +1752,16 @@ int spx_primal(glp_prob *P, const glp_smcp *parm)
       csa->tm_beg = xtime();
       csa->it_beg = csa->it_cnt = P->it_cnt;
       csa->it_dpy = -1;
+#if 1 /* 15/VII-2017 */
+      csa->tm_dpy = 0.0;
+#endif
       csa->inv_cnt = 0;
+#if 1 /* 01/VII-2017 */
+      csa->degen = 0;
+#endif
+#if 1 /* 23/VI-2017 */
+      csa->ns_cnt = csa->ls_cnt = 0;
+#endif
       /* try to solve working LP */
       ret = primal_simplex(csa);
       /* return basis factorization back to problem object */
@@ -1402,11 +1780,26 @@ int spx_primal(glp_prob *P, const glp_smcp *parm)
       spx_store_basis(csa->lp, P, map, daeh);
       /* compute simplex multipliers for final basic solution found by
        * the solver */
+#if 0 /* 09/VII-2017 */
       spx_eval_pi(csa->lp, csa->work);
+#else
+      spx_eval_pi(csa->lp, csa->work.vec);
+#endif
       /* convert working LP solution to original LP solution and store
-       * it to problem object */
+       * it into the problem object */
+#if SCALE_Z
+      for (i = 1; i <= csa->lp->m; i++)
+         csa->work.vec[i] *= csa->fz;
+      for (j = 1; j <= csa->lp->n-csa->lp->m; j++)
+         csa->d[j] *= csa->fz;
+#endif
+#if 0 /* 09/VII-2017 */
       spx_store_sol(csa->lp, P, SHIFT, map, daeh, csa->beta, csa->work,
          csa->d);
+#else
+      spx_store_sol(csa->lp, P, parm->shift, map, daeh, csa->beta,
+         csa->work.vec, csa->d);
+#endif
       tfree(daeh);
       /* save simplex iteration count */
       P->it_cnt = csa->it_cnt;
@@ -1431,7 +1824,7 @@ skip: /* deallocate working objects and arrays */
       spx_free_lp(csa->lp);
       tfree(map);
       tfree(csa->orig_c);
-#if PERTURB
+#if 1 /*PERTURB*/
       tfree(csa->orig_l);
       tfree(csa->orig_u);
 #endif
@@ -1451,7 +1844,15 @@ skip: /* deallocate working objects and arrays */
       fvs_free_vec(&csa->tcol);
       fvs_free_vec(&csa->trow);
 #endif
+#if 1 /* 23/VI-2017 */
+      if (csa->bp != NULL)
+         tfree(csa->bp);
+#endif
+#if 0 /* 09/VII-2017 */
       tfree(csa->work);
+#else
+      fvs_free_vec(&csa->work);
+#endif
       /* return to calling program */
       return ret;
 }
